#lang racket/base
(require racket/list
         parser-tools/cfg-parser
         parser-tools/lex
         (for-syntax racket/base
                     syntax/parse)
         syntax/readerr
         "lexer.rkt")

(provide parse)

(define (parse src-name in)
  (parameterize ([current-source src-name])
    (parse-from-lex (lambda () (lex in)))))

(define parse-from-lex
  (cfg-parser
   (start <program>)
   (end EOF)
   (src-pos)
   (tokens content-tokens
           delimiter-tokens)
   (precs)
   (error (lambda (a t v start end) 
            (raise-parse-error t v start end)))
   (grammar 
    (<program> [(<ws> <classes> <expr> <ws> SEMICOLON <ws>)
                `(,@$2 ,$3)])
    (<classes> [() empty]
               [(<class> <ws> <classes>) (cons $1 $3)])
    (<class> [(CLASS <ws> ID <ws> EXTENDS <ws> ID <ws> OPENB <ws>
                     <fields>
                     <methods> 
                     CLOSEB)
              (->stx `(class ,$3 extends ,$7 ,$11 ,@$12))])
    (<fields> [() empty]
              [(<field> <ws> <fields>) (cons $1 $3)])
    (<field> [(<type> <ws> ID <ws> SEMICOLON) (->stx `[,$3 : ,$1])])
    (<methods> [() empty]
               [(<method> <ws> <methods>) (cons $1 $3)])
    (<method> [(<type> <ws> ID <ws>
                       OPENP <ws> <type> <ws> ARG <ws> CLOSEP
                       <ws> OPENB
                       <ws> RETURN <ws> <expr> <ws> SEMICOLON
                       <ws> CLOSEB)
               (->stx `{,$3 : ,$7 -> ,$1 ,$17})])
    (<type> [(NUM) (->stx 'num)]
            [(ID) (->stx $1)])
    (<expr> [(<texpr> <ws> PLUS <ws> <expr>) (->stx `{+ ,$1 ,$5})]
            [(<texpr>) $1])
    (<texpr> [(<oexpr> <ws> TIMES <ws> <texpr>) (->stx `{* ,$1 ,$5})]
             [(<oexpr>) $1])
    (<oexpr> [(THIS) (->stx 'this)]
             [(ARG) (->stx 'arg)]
             [(NUMBER) (->stx $1)]
             [(OPENP <ws> <expr> <ws> CLOSEP) $3]
             [(<oexpr> <ws> DOT <ws> ID)
              (->stx `{get ,$1 ,$5})]
             [(<oexpr> <ws> DOT <ws> ID <ws> 
                       OPENP <ws> <expr> <ws> CLOSEP)
              (->stx `{send ,$1 ,$5 ,$9})]
             [(SUPER <ws> DOT <ws> ID <ws>
                     OPENP <ws> <expr> <ws> CLOSEP)
              (->stx `{super ,$5 ,$9})]
             [(NEW <ws> ID <ws> OPENP <ws> <exprs> CLOSEP)
              (->stx `{new ,$3 ,@$7})])
    (<exprs> [(<ws>) empty]
             [(<exprs+> <ws>) $1])
    (<exprs+> [(<expr>) (list $1)]
              [(<expr> <ws> COMMA <ws> <exprs+>) (cons $1 $5)])
    (<ws> [() #f]
          [(WHITESPACE) #f]))))
              

(define-syntax (->stx stx)
  ;; A non-hygienic marco to access $1-start-pos and $n-end-pos, which
  ;; are bound by the `cfg-parse` form in a grammar production's
  ;; action:
  (syntax-parse stx
    [(_ e)
     (let ([start (datum->syntax stx '$1-start-pos)]
           [end (datum->syntax stx '$n-end-pos)])
       #`(datum->syntax #f e (to-srcloc #,start #,end) orig-prop))]))

;; This property tells DrRacket that a constructed syntax object
;; should be treated as being in the original source:
(define orig-prop (read-syntax 'src (open-input-bytes #"x")))

;; ----------------------------------------
;; Source locations and error reporting:

(define current-source (make-parameter #f))

(define (to-srcloc start end)
  ;; Convert parser source locations to syntax-object
  ;; and `raise-read-error` format:
  (list
   (current-source)
   (position-line start)
   (position-col start)
   (position-offset start)
   (and (position-offset end)
        (position-offset start)
        (- (position-offset end)
           (position-offset start)))))

(define (raise-parse-error t v start end)
  (apply
   (if (eq? t 'EOF) raise-read-eof-error raise-read-error) 
   (format "bad syntax at ~a" (token->string t v))
   (to-srcloc start end)))
